---------------
-- FUNCTIONS --
---------------

-- Spawn a tether or laser from one mobj to another.
local function spawnVine(x1, y1, z1, x2, y2, z2, color, mo)
	mo.vinemobjs = $ or {}

	local ox, oy, oz, thok, frame
	for i = 1, 35 do
		ox = FixedMul(i*((x2-x1)/35), mo.scale)
		oy = FixedMul(i*((y2-y1)/35), mo.scale)
		oz = FixedMul(i*((z2-z1)/35), mo.scale)

		if mo.vinemobjs[i] and mo.vinemobjs[i].valid then
			P_MoveOrigin(mo.vinemobjs[i], x1+ox, y1+oy, z1+oz)
		else
			thok = P_SpawnMobj(x1+ox, y1+oy, z1+oz, MT_DUMMYOBJ)
			thok.scale = (3*FRACUNIT/2) + FRACUNIT*cos(i*ANGLE_90)
			thok.color = color or SKINCOLOR_WHITE
			thok.sprite = SPR_LSTM
			thok.scale = mo.scale

			frame = H
			if (i == 6 or i == 20) then
				frame = F
			elseif (i == 15 or i == 18) then
				frame = I
			elseif (i == 4 or i == 24 or i == 32) then
				frame = G
			end

			thok.frame = frame
			thok.fuse = 38
			thok.tics = 38
			mo.vinemobjs[i] = thok
		end
	end
end

-- spawnVine but accept mobjs.
local function spawnVineFromMobj(mo, mo2, color)
	if not mo or not mo.valid then return end
	if not mo2 or not mo2.valid then return end

	spawnVine(mo.x, mo.y, mo.z+(mo.height/2), mo2.x, mo2.y, mo2.z, color or SKINCOLOR_WHITE, mo)
end

----------------
-- MAIN STUFF --
----------------
addHook("AbilitySpecial", function(p)
	if not p.mo or not p.mo.valid then return end
	if p.mo.skin ~= "blossom" then return end
	if p.mo.blossom_tornado then return end
	if p.mo.blossom_didvine and not p.powers[pw_super] then return end
	if p.mo.blossom_fastfall then return end
	if p.mo.blossom_twirltimer then return end
	if p.exiting then return true end
	if p.mo.targvine then return end
	p.mo.blossom_set_targvine = nil
	p.mo.blossom_addmom = nil

	p.mo.state = S_PLAY_BLOSSOM_TORNADO
	S_StartSound(p.mo, sfx_grass1)

	local vine = P_SpawnMobjFromMobj(p.mo, 0, 0, p.mo.height, MT_VINETARG)
	vine.color = p.skincolor
	vine.targoffset = 2
	vine.user = p

	local enm = P_LookForEnemies(p, true)
	if enm and enm.valid and not enm.player then
		p.mo.targvine = {vine, p.mo.angle, 35, enm, p.speed}
		p.mo.blossom_didtornado = nil
		p.mo.blossom_didvine = nil
	else
		p.mo.blossom_didvine = true
		p.mo.targvine = {vine, p.mo.angle, 35}
		P_SetObjectMomZ(p.mo, 5*FRACUNIT)
	end
end)

rawset(_G, "Blossom_Vines", function(p)
	if not p.mo or not p.mo.valid then return end
	if p.mo.skin ~= "blossom" then return end

	-- Super Boost Particles
	if p.mo.blossom_didsuper then
		p.mo.blossom_didsuper = $-1
		if p.mo.blossom_didsuper <= 0 then p.mo.blossom_didsuper = nil end

		if not (leveltime%12) then
			if abs(p.mo.momz) < p.speed then
				p.mo.blossom_didsuper_horizontal = p.mo.blossom_didsuper
				p.mo.blossom_didsuper = nil
			else
				local j = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_DUMMYOBJ)
				j.sprite = SPR_BLBS
				j.frame = FF_FULLBRIGHT|FF_TRANS50|A
				j.fuse = TICRATE
				j.scale = p.mo.scale
				j.destscale = 5*p.mo.scale
				j.momz = p.mo.momz/2
				j.renderflags = RF_SLOPESPLAT|RF_NOSPLATBILLBOARD
				j.flags2 = MF2_SPLAT

				P_CreateFloorSpriteSlope(j)
				j.floorspriteslope.o = {
					x = j.x,
					y = j.y,
					z = j.z
				}
			end
		end
	end

	if p.mo.blossom_didsuper_horizontal then
		p.mo.blossom_didsuper_horizontal = $-1
		if p.mo.blossom_didsuper_horizontal <= 0 then p.mo.blossom_didsuper_horizontal = nil end

		if not (leveltime%9) then
			if abs(p.mo.momz) > p.speed then
				p.mo.blossom_didsuper = p.mo.blossom_didsuper_horizontal
				p.mo.blossom_didsuper_horizontal = nil
			else
				local j = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_DUMMYOBJ)
				j.sprite = SPR_BLBS
				j.frame = FF_FULLBRIGHT|FF_TRANS50|FF_PAPERSPRITE|A
				j.fuse = TICRATE
				j.scale = p.mo.scale
				j.destscale = 5*p.mo.scale
				j.angle = p.mo.angle+ANGLE_90
			end
		end
	end

	-- Spring Flag
	if p.mo.blossom_autospringremove then
		p.mo.blossom_autospringremove = $-1
		if p.mo.blossom_autospringremove <= 0 then
			p.powers[pw_strong] = $ & ~STR_SPRING
			p.mo.blossom_autospringremove = nil
		end
	end

	-- Dive
	if p.mo.blossom_didvine and not p.mo.targvine and p.mo.buttons.jump > 0 and p.mo.buttons.spin > 0 and not p.mo.blossom_fastfall and not (p.mo.eflags & MFE_GOOWATER) then
		p.drawangle = p.mo.angle
		p.mo.blossom_fastfall = true
		p.pflags = $ & ~PF_SPINNING
		p.mo.blossom_grounded = nil
		p.mo.blossom_tornado = nil
		p.mo.blossom_cancelled = nil
		p.mo.blossom_slidetimer = nil
		p.pflags = $ & ~PF_CANCARRY
		S_StartSound(p.mo, sfx_s3k51)
	end

	-- Wall
	if p.mo.blossom_wallslide then
		if p.mo.blossom_superjump then
			p.mo.blossom_superjump = $-1
			if p.mo.blossom_superjump <= 0 then p.mo.blossom_superjump = nil end
		end

		local line = p.mo.blossom_wallslide
		local x, y = P_ClosestPointOnLine(p.mo.x, p.mo.y, line)
		local linedist = R_PointToDist2(p.mo.x, p.mo.y, x, y)
		local fangle = R_PointToAngle2(p.mo.x, p.mo.y, x, y) + FixedAngle(180*FRACUNIT)
		p.drawangle = fangle
		P_SlideMove(p.mo)

		p.powers[pw_nocontrol] = 32768

		p.mo.momx = $+5*cos(fangle + FixedAngle(180*FRACUNIT))
		p.mo.momy = $+5*sin(fangle + FixedAngle(180*FRACUNIT))
		p.mo.momz = max(-5*p.mo.scale, $-p.mo.scale/3)

		p.mo.flags = $|MF_NOGRAVITY
		p.mo.state = S_PLAY_CLING

		-- Walljump sound.
		if not S_SoundPlaying(p.mo, sfx_cdfm70) then S_StartSound(p.mo, sfx_cdfm70) end

		-- Dust
		if not (leveltime%6) then
			local dust = P_SpawnMobjFromMobj(p.mo, P_RandomRange(-16, 16)*FRACUNIT, P_RandomRange(-16, 16)*FRACUNIT, p.mo.height, MT_SPINDUST)
			dust.scale = 2*p.mo.scale/3
			P_SetObjectMomZ(dust, P_RandomRange(2, 5)*FRACUNIT)
		end

		-- Jump
		if p.mo.buttons.jump == 1 then
			if p.mo.blossom_superjump then
				p.powers[pw_nocontrol] = 0
				S_StartSound(p.mo, sfx_bjmpsp)
				P_InstaThrust(p.mo, fangle, 10*p.mo.scale)
				P_SetObjectMomZ(p.mo, 13*FRACUNIT, true)
				p.drawangle = fangle

				if p.mo.blossom_addmom then
					p.mo.momx = $ + p.mo.blossom_addmom[1]
					p.mo.momy = $ + p.mo.blossom_addmom[2]
				end

				p.mo.blossom_addmom = nil
				p.mo.blossom_superjump = nil
				p.mo.blossom_didsuper = 35
			else
				p.powers[pw_nocontrol] = 0
				S_StartSound(p.mo, sfx_ngjump)
				P_InstaThrust(p.mo, fangle, 9*p.mo.scale)
				P_SetObjectMomZ(p.mo, 8*p.mo.scale, true)
				p.drawangle = fangle
			end

			p.mo.flags = $ & ~MF_NOGRAVITY
			p.mo.blossom_wallslide = nil
			p.mo.blossom_slidetimer = nil
			p.mo.state = S_PLAY_SPRING
		elseif p.mo.buttons.spin == 1 and not (p.mo.eflags & MFE_GOOWATER) then
			p.powers[pw_nocontrol] = 0
			S_StartSound(p.mo, sfx_jump)
			P_InstaThrust(p.mo, fangle, 9*p.mo.scale)
			P_SetObjectMomZ(p.mo, 8*p.mo.scale, true)
			p.mo.blossom_didvine = nil

			p.mo.flags = $ & ~MF_NOGRAVITY
			p.pflags = $ & ~PF_SPINNING
			p.pflags = $ & ~PF_CANCARRY
			p.mo.blossom_wallslide = nil
			p.drawangle = p.mo.angle
			p.mo.blossom_fastfall = true
			p.mo.blossom_slidetimer = nil
			S_StartSound(p.mo, sfx_s3k51)
		end

		-- Stop if we touch the ground
		if P_IsObjectOnGround(p.mo) then
			p.powers[pw_nocontrol] = 0
			p.mo.blossom_notonwall = nil
			p.mo.flags = $ & ~MF_NOGRAVITY
			p.mo.blossom_wallslide = nil
			p.mo.blossom_slidetimer = nil
			p.mo.state = S_PLAY_STND
		end

		-- lets also try this
		if not p.powers[pw_pushing] or linedist >= 64*p.mo.scale then
			p.mo.blossom_notonwall = $ and $+1 or 1
			if p.mo.blossom_notonwall >= 6 then
				p.powers[pw_nocontrol] = 0
				p.mo.blossom_notonwall = nil
				p.mo.flags = $ & ~MF_NOGRAVITY
				p.mo.blossom_wallslide = nil
				p.mo.blossom_slidetimer = nil
				p.mo.state = S_PLAY_SPRING
			end
		else
			p.mo.blossom_notonwall = nil
		end

		--- can i like kill myself
		if (line.frontsector)
			if ((line.frontsector.ceilingheight == line.frontsector.floorheight) and (p.mo.z > line.frontsector.ceilingheight))
			or (line.backsector and line.frontsector.floorheight > line.backsector.floorheight and p.mo.z > line.frontsector.floorheight)
			or (line.backsector and line.frontsector.ceilingheight < line.backsector.ceilingheight and (p.mo.z+p.mo.height) > line.frontsector.ceilingheight)
				p.mo.flags = $ & ~MF_NOGRAVITY
				p.mo.blossom_wallslide = nil
				p.mo.blossom_slidetimer = nil
				p.mo.state = S_PLAY_FALL
				p.powers[pw_nocontrol] = 0
			end
		end

		if (line.backsector)
			if ((line.backsector.ceilingheight == line.backsector.floorheight) and (p.mo.z > line.backsector.ceilingheight))
			or (line.frontsector and line.backsector.floorheight > line.frontsector.floorheight and p.mo.z > line.backsector.floorheight)
			or (line.frontsector and line.backsector.ceilingheight < line.frontsector.ceilingheight and (p.mo.z+p.mo.height) > line.backsector.ceilingheight)
				p.mo.flags = $ & ~MF_NOGRAVITY
				p.mo.blossom_wallslide = nil
				p.mo.blossom_slidetimer = nil
				p.mo.state = S_PLAY_FALL
				p.powers[pw_nocontrol] = 0
			end
		end

		if p.mo.momz >= 0 then p.mo.blossom_slidetimer = 0 end
		p.mo.blossom_slidetimer = $ and $+1 or 1
		if p.mo.blossom_slidetimer >= 30 then
			p.mo.state = S_PLAY_PAIN
			p.powers[pw_nocontrol] = 0
			p.mo.blossom_notonwall = nil
			p.mo.flags = $ & ~MF_NOGRAVITY
			p.mo.blossom_wallslide = nil
			p.mo.blossom_slidetimer = nil
			p.mo.state = S_PLAY_SPRING
		end
	end

	-- Vine cling
	if p.mo.targvine then
		local vine = p.mo.targvine[1]

		p.mo.backup = $ and $+1 or 1
		if p.mo.backup >= 5*TICRATE then
			if vine and vine.valid then P_RemoveMobj(vine) end
			p.mo.targvine = nil
			p.mo.backup = nil
			p.mo.targvine = nil
			return
		end

		p.drawangle = p.mo.targvine[2]
		p.mo.targvine[3] = $-1

		-- Time
		local time = p.mo.targvine[3]

		-- Kara Dive
		if time >= 25 and p.mo.buttons.spin == 1 and not (p.mo.eflags & MFE_GOOWATER) then
			p.mo.blossom_fastfall = true

			if time <= 30 and (p.mo.targvine[4] or vine.hitwall) then
				p.mo.state = S_PLAY_BLOSSOM_TORNADO
				p.mo.blossom_didsuper_horizontal = 30
				if p.cmd.forwardmove >= 50 then P_InstaThrust(p.mo, p.mo.angle, 15*FRACUNIT) end
				S_StartSound(p.mo, sfx_bjmpsp)
				S_StartSound(p.mo, sfx_s3k51)
			end

			if vine and vine.valid then P_RemoveMobj(vine) end
			p.mo.targvine = nil
			p.mo.backup = nil
			return
		end

		if time > 20 then
			vine.targoffset = $+((256-$)/10)
		else
			vine.targoffset = $-($/10)
		end

		if p.mo.targvine[4] then
			local mobj = p.mo.targvine[4]
			p.mo.blossom_invinc = 10

			if mobj and mobj.valid then
				vine.momx = (mobj.x-vine.x)/2
				vine.momy = (mobj.y-vine.y)/2
				vine.momz = (mobj.z-vine.z)/2

				if time > 30 then
				elseif time > 26 then
					p.mo.momx = 0
					p.mo.momy = 0
					p.mo.momz = 0
				else
					local dist = P_AproxDistance(P_AproxDistance(mobj.x - p.mo.x, mobj.y - p.mo.y), mobj.z - p.mo.z)
					p.mo.momx = FixedMul(FixedDiv(mobj.x - p.mo.x, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))
					p.mo.momy = FixedMul(FixedDiv(mobj.y - p.mo.y, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))
					p.mo.momz = FixedMul(FixedDiv(mobj.z - p.mo.z, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))

					if (p.cmd.buttons & BT_SPIN) and not (mobj.flags & MF_BOSS) then
						p.mo.state = S_PLAY_BLOSSOM_TORNADO
						p.mo.blossom_didsuper_horizontal = 30
						P_InstaThrust(p.mo, p.mo.angle, p.mo.blossom_vinespeed or 20*FRACUNIT)
						P_SetObjectMomZ(p.mo, 8*p.mo.scale)
						S_StartSound(p.mo, sfx_bjmpsp)
						S_StartSound(p.mo, sfx_s3k80)
						P_DamageMobj(mobj, p.mo)
						p.mo.targvine = nil
					end

					if (dist <= 80*FRACUNIT) then
						if (mobj.flags & MF_BOSS) then
							P_InstaThrust(p.mo, p.drawangle, -10*FRACUNIT)
							P_SetObjectMomZ(p.mo, 14*FRACUNIT)
						elseif not (mobj.flags & MF_SPRING) then
							P_InstaThrust(p.mo, p.drawangle, p.mo.blossom_vinespeed or 40*FRACUNIT)
							P_SetObjectMomZ(p.mo, 10*FRACUNIT)
						end

						if (mobj.flags & MF_ENEMY or mobj.flags & MF_BOSS)
							P_DamageMobj(mobj, p.mo, p.mo)
							p.mo.state = S_PLAY_BLOSSOM_TORNADO
							p.pflags = $|PF_JUMPED
							p.mo.targvine = nil
						elseif (mobj.flags & MF_MONITOR) then
							P_KillMobj(mobj, p.mo, p.mo)
							p.mo.state = S_PLAY_BLOSSOM_TORNADO
							p.pflags = $|PF_JUMPED
							p.mo.targvine = nil
						elseif (mobj.flags & MF_SPRING) then
							p.mo.blossom_autospringremove = 5
							p.powers[pw_strong] = STR_SPRING
							p.mo.targvine = nil
						end
					end
				end
			else
				S_StartSound(p.mo, sfx_s3k4a)
				P_SetObjectMomZ(p.mo, 10*FRACUNIT)
				if vine and vine.valid then P_RemoveMobj(vine) end
				p.mo.targvine = nil
			end
		elseif vine.hitwall then
			vine.momx = 0
			vine.momy = 0
			vine.momz = 0

			if time > 30 then
			elseif time > 20 then
				p.mo.momx = 0
				p.mo.momy = 0
				p.mo.momz = 0
			else
				p.drawangle = R_PointToDist2(p.mo.x, p.mo.y, vine.x, vine.y)

				local dist = P_AproxDistance(P_AproxDistance(vine.x - p.mo.x, vine.y - p.mo.y), vine.z - p.mo.z)
				p.mo.momx = FixedMul(FixedDiv(vine.x - p.mo.x, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))
				p.mo.momy = FixedMul(FixedDiv(vine.y - p.mo.y, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))
				p.mo.momz = FixedMul(FixedDiv(vine.z - p.mo.z, dist), min(80*FRACUNIT, p.mo.blossom_vinespeed or 80*FRACUNIT))
				p.mo.blossom_addmom = {p.mo.momx, p.mo.momy}

				if (p.cmd.buttons & BT_SPIN) then
					p.mo.state = S_PLAY_BLOSSOM_TORNADO
					p.mo.blossom_didsuper_horizontal = 30
					P_InstaThrust(p.mo, p.mo.angle, p.mo.blossom_vinespeed)
					P_SetObjectMomZ(p.mo, 8*FRACUNIT)
					S_StartSound(p.mo, sfx_bjmpsp)
					S_StartSound(p.mo, sfx_s3k80)
					p.mo.targvine = nil
				end

				if (dist <= 60*FRACUNIT) then
					S_StartSound(p.mo, sfx_s3k4a)
					P_SetObjectMomZ(p.mo, 10*FRACUNIT)
					p.mo.blossom_wallslide = vine.hitwall
					p.mo.blossom_superjump = 6
					if vine and vine.valid then P_RemoveMobj(vine) end
					p.mo.targvine = nil
				end
			end
		elseif p.mo.gotovine then
			vine.momx = (p.mo.x-vine.x)/3
			vine.momy = (p.mo.y-vine.y)/3
			vine.momz = (p.mo.z-vine.z)/3

			if p.mo.blossom_vinestorage then
				p.mo.blossom_vinestorage = nil
				p.mo.momx = 0
				p.mo.momy = 0
				p.mo.momz = max(50*FRACUNIT, p.speed)*P_MobjFlip(p.mo)

				p.mo.gotovine = nil
				p.mo.targvine = nil
				p.mo.state = S_PLAY_BLOSSOM_TORNADO
				S_StartSound(p.mo, sfx_bjmpsp)
				p.mo.blossom_didsuper = 30
				if vine and vine.valid then P_RemoveMobj(vine) end
				p.mo.momz = 2*$/3
			else
				p.mo.momx = p.mo.gotovine[1]
				p.mo.momy = p.mo.gotovine[2]
				p.mo.momz = p.mo.gotovine[3]
				p.mo.gotovine[4] = $-1

				if (p.mo.gotovine[4] <= 0) then
					p.mo.gotovine = nil
					p.mo.state = S_PLAY_BLOSSOM_TORNADO
					S_StartSound(p.mo, sfx_bjmpsp)
					p.mo.blossom_didsuper_horizontal = 30
					if vine and vine.valid then P_RemoveMobj(vine) end
					p.mo.targvine = nil
					p.mo.momz = 2*$/3
				end
			end
		else
			p.mo.blossom_vinespeed = max(52*p.mo.scale, p.speed) or 52*p.mo.scale

			local x = p.mo.x + FixedMul(vine.targoffset*cos(p.drawangle)*3, p.mo.scale)
			local y = p.mo.y + FixedMul(vine.targoffset*sin(p.drawangle)*3, p.mo.scale)
			local z = p.mo.z + FixedMul(vine.targoffset*FRACUNIT, p.mo.scale)*P_MobjFlip(p.mo)
			vine.momx = ((x-vine.x)/3) + p.mo.momx
			vine.momy = ((y-vine.y)/3) + p.mo.momy
			vine.momz = ((z-vine.z)/3) + p.mo.momz

			if (p.mo.buttons.jump == 1) and (time <= 20) then
				local dist = P_AproxDistance(P_AproxDistance(vine.x - p.mo.x, vine.y - p.mo.y), vine.z - p.mo.z)
				local xspd = FixedMul(FixedDiv(vine.x - p.mo.x, dist), p.mo.blossom_vinespeed)
				local yspd = FixedMul(FixedDiv(vine.y - p.mo.y, dist), p.mo.blossom_vinespeed)
				local zspd = FixedMul(FixedDiv(vine.z - p.mo.z, dist), p.mo.blossom_vinespeed)

				p.mo.gotovine = {xspd, yspd, zspd, max(5, 10 - FixedInt(p.mo.blossom_vinespeed/4))}
				S_StartSound(p.mo, sfx_ngjump)
			end
		end

		-- test
		spawnVineFromMobj(p.mo, vine, p.powers[pw_super] and p.mo.color or p.skincolor)

		if p.mo.blossom_wallslide then
			if vine and vine.valid then P_RemoveMobj(vine) end
			p.mo.targvine = nil
		end

		if time <= 0 then p.mo.targvine = nil end
	else
		if p.mo.gotovine and not (leveltime%2) then p.mo.blossom_vinestorage = true end
	end

	-- Target locking
	if not p.mo.blossom_didvine and not p.mo.blossom_tornado and not P_IsObjectOnGround(p.mo) and P_LookForEnemies(p, true)
		local targ = P_LookForEnemies(p, true)
		if targ.player then return end

		P_SpawnLockOn(p, targ, S_LOCKON3)
	end
end)